require("version.nut");
require("town.nut");

/* Import SuperLib for GameScript */
import("util.superlib", "SuperLib", 27);
Result <- SuperLib.Result;
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;
ScoreList <- SuperLib.ScoreList;
Tile <- SuperLib.Tile;
Direction <- SuperLib.Direction;
Town <- SuperLib.Town;
Industry <- SuperLib.Industry;

class MainClass extends GSController 
{
	constructor()
	{
	}

	towns = null;
	last_date = 0;
	current_year = 0;
	last_master_switch_value = true;

	loaded_data = null; // assigned when loading a save game
	loaded_from_version = null; // assigned when loading a save game

	function Save();
	function Load(version, data);
}

function MainClass::Start()
{
	local start_tick = GSController.GetTick();
	this.Init();
	local setup_duration = GSController.GetTick() - start_tick;

	Log.Info("Game setup done", Log.LVL_INFO);
	Log.Info("Setup took " + setup_duration + " ticks", Log.LVL_INFO);
	Log.Info("", Log.LVL_INFO);

	Log.Info("Happy playing", Log.LVL_INFO);

	GSController.Sleep(1);

	this.ShowUpdateNotice();

	while (true) {
		this.HandleMasterSwitch();
		this.HandleEvents();
		if(this.last_master_switch_value == true)
		{
			this.ManageTowns();
		}

		this.Sleep(60);
	}
}

function MainClass::Init()
{
	this.last_master_switch_value = GSController.GetSetting("master_switch") == 1;
	if(this.last_master_switch_value)
	{
		GSGameSettings.SetValue("economy.town_growth_rate", 2);
		GSGameSettings.SetValue("economy.fund_buildings", 0);
	}
	this.last_date = GSDate.GetCurrentDate();
	this.current_year = GSDate.GetYear(this.last_date);
	Log.Info("Init: master switch: " + this.last_master_switch_value, Log.LVL_INFO);

	// Create the towns list
	Log.Info("Create town list ... (can take a while on large maps)", Log.LVL_INFO);
	this.towns = this.CreateTownList();

	// Set initial time the towns will grow
	InitalizeManageTimers(this.towns , this.last_master_switch_value);
}

// Non-class function
function InitalizeManageTimers(town_list, manage = true)
{
	foreach (town in town_list) {
		town.check_timer = GSBase.RandRange(20) + 10;
		if (manage)
			town.ManageTown();
	}
}

function MainClass::HandleEvents()
{
	if (GSEventController.IsEventWaiting())
	{
		local ev = GSEventController.GetNextEvent();

		if (ev == null)
			return;

		
		local ev_type = ev.GetEventType();
		if(ev_type == GSEvent.ET_TOWN_FOUNDED)
		{
			// Trigger updating of town list directly when the
			// town event is received. (eg. don't wait for the
			// periodical town update)
			if(this.last_master_switch_value == true)
				this.CheckForNewTowns();
		}
	}
}

function MainClass::CreateTownList()
{
	local result = [];

	/* Make a squirrel array of all towns.
	 * For each town, an instance of GoalTown is created
	 * to store the data related to that town
	 */
	local all_towns = GSTownList();
	foreach (t, _ in all_towns)
	{
		result.append(GoalTown(t));
	}

	return result;
}

function MainClass::CheckForNewTowns()
{
	// Do our town-list have fewer towns than the game?
	local world_towns = GSTownList();
	if (this.towns.len() < world_towns.Count())
	{
		// Create a GSList with all towns that we have
		local have_towns = GSList();
		foreach (town in this.towns)
		{
			have_towns.AddItem(town.id, 0);
		}
		
		// Reduce the world list of towns with those that we already have
		world_towns.RemoveList(have_towns);

		// Add the remaining towns to our list
		// Also, keep a list of newly added towns
		local new_towns = [];
		foreach (town_id, _ in world_towns)
		{
			local town = GoalTown(town_id);
			this.towns.append(town);
			new_towns.append(town);
		}

		// Perform an initial manage of the new towns
		InitalizeManageTimers(new_towns, true);
	}
}

function MainClass::ManageTowns()
{
	local date = GSDate.GetCurrentDate();
	local diff = date - this.last_date;
	this.last_date = date;

	if (diff == 0) return;

	this.CheckForNewTowns();

	foreach (town in this.towns) {

		town.check_timer -= diff;

		if (town.check_timer <= 0) {
			town.check_timer = town.check_timer + GSBase.RandRange(20) + 10;
			town.ManageTown();
		}
	}
}

function MainClass::HandleMasterSwitch()
{
	local new_value = GSController.GetSetting("master_switch") == 1;

	// Have the master switch changed its value?
	if (new_value != this.last_master_switch_value)
	{
		Log.Info("Detected change of master switch", Log.LVL_INFO);

		if (new_value == true)
		{
			// GS have been re-enabled
			Log.Info("> Enable Neighbours are important :-)", Log.LVL_INFO);

			GSGameSettings.SetValue("economy.town_growth_rate", 2);
			GSGameSettings.SetValue("economy.fund_buildings", 0);

			// Any NewTown events that occurred while the GS have been disabled
			// have been thrown away. So check if there are any new towns.
			this.CheckForNewTowns();

			this.ManageTowns();
			Log.Info("> The script have been re-enabled :-)", Log.LVL_INFO);
		}
		else
		{
			// GS have been disabled - restore default town growth
			Log.Info("> Disable Neighbours are important", Log.LVL_INFO);

			// Show a notice to allow some time for towns to restore
			for (local companyID = GSCompany.COMPANY_FIRST; companyID <= GSCompany.COMPANY_LAST; companyID++) {
				if (GSCompany.ResolveCompanyID(companyID) != GSCompany.COMPANY_INVALID) {
					GSGoal.Question(1, companyID, GSText(GSText.STR_NOTICE_DISABLING_TAKES_TIME), GSGoal.QT_WARNING, GSGoal.BUTTON_CLOSE);
				}
			}

			foreach (town in this.towns) {
				town.Restore();
			}

			GSGameSettings.SetValue("economy.town_growth_rate", 2);
			GSGameSettings.SetValue("economy.fund_buildings", 1);

			Log.Info("> Completed restoring towns + game settings. (now allow time for OpenTTD to manage all towns)", Log.LVL_INFO);
		}

		this.last_master_switch_value = new_value;
	}
}

function MainClass::ShowUpdateNotice()
{
	// Don't show update notices for new games
	if(this.loaded_data == null) return;

	if(this.loaded_from_version < 6)
	{
		// Loaded from a version that used the old legacy method for
		// comparing town sizes. Version 1-5 didn't have the setting
		// introduced in version 6, so the game have now defaulted to
		// the new method. Display a notice to players about this.
		for (local companyID = GSCompany.COMPANY_FIRST; companyID <= GSCompany.COMPANY_LAST; companyID++) {
			if (GSCompany.ResolveCompanyID(companyID) != GSCompany.COMPANY_INVALID) {
				GSGoal.Question(0, companyID, GSText(GSText.STR_NOTICE_NEW_TOWN_COMPARE_METHOD), GSGoal.QT_WARNING, GSGoal.BUTTON_CLOSE);
			}
		}
	}

}

function MainClass::Save()
{
	Log.Info("Saving ... (no data is saved)", Log.LVL_SUB_DECISIONS);
	local result = {};
	return result;
}

function MainClass::Load(version, table)
{
	Log.Info("Loading..", Log.LVL_INFO);
	Log.Info("Previously saved with Game Script version " + version, Log.LVL_INFO);

	if(version > SELF_VERSION)
	{
		Log.Warning("Warning: Loading from a newer version of Neighbours are important", Log.LVL_INFO);
	}
	this.loaded_data = table;
	this.loaded_from_version = version;
}
